--
-- Author: Sergei Jukov(sjukov@staff.total-knowledge.com)
-- Modified by: Ilya Volynets(ilya@selfsoft.com)
--
CREATE OR REPLACE PACKAGE owa_cookie AS
    TYPE vc_arr IS TABLE OF VARCHAR2(4000); -- INDEX BY BINARY_INTEGER;				   
    TYPE cookie IS RECORD(
	name VARCHAR2(4000),
	vals VC_ARR,
	num_vals INTEGER
    );

    PROCEDURE send(	name IN VARCHAR2,
					value IN VARCHAR2,
					expires IN DATE DEFAULT NULL,
					path IN VARCHAR2 DEFAULT NULL,
					domain IN VARCHAR2 DEFAULT NULL,
					secure IN VARCHAR2 DEFAULT NULL);
    PROCEDURE remove(name IN VARCHAR2,
					 value IN VARCHAR2,
					 path IN VARCHAR2 DEFAULT NULL);
    PROCEDURE get_all(names out VC_ARR,
				   vals out VC_ARR,
				    num_vals out INTEGER);    
    FUNCTION get(name VARCHAR2) RETURN cookie;				
END;
/
show err

----------------------------
CREATE OR REPLACE PACKAGE BODY owa_cookie AS
    PROCEDURE  send(name IN VARCHAR2,
		    value IN VARCHAR2,
		    expires IN DATE DEFAULT NULL,
		    path IN VARCHAR2 DEFAULT NULL,
		    domain IN VARCHAR2 DEFAULT NULL,
		    secure IN VARCHAR2 DEFAULT NULL)
	AS
		string_cookies VARCHAR2(4000);
		cur_date DATE;
		max_age NUMBER;
	BEGIN
--	    string_cookies:='Set-Cookie:'||name||'='||value;
	    string_cookies:=name||'='||value;
	    IF expires IS NOT NULL THEN
		cur_date:=CURRENT_DATE;
		max_age:=expires-cur_date;
		string_cookies:=string_cookies||'; Max-Age='||max_age;
	    END IF;
	    IF path IS NOT NULL THEN
			string_cookies:=string_cookies||'; path='||path;
	    END IF;
	    IF domain IS NOT NULL THEN
			string_cookies:=string_cookies||'; domain='||domain;
	    END IF;
	    IF secure IS NOT NULL THEN
			string_cookies:=string_cookies||'; secure';
	    END IF;
	    string_cookies:=string_cookies||'; version=1';
	    --oci_www.appendHeader('Set-Cookie',string_cookies);
	    oci_www.putStringHdr('Set-Cookie: '||string_cookies);
	END;
----------------------------
    PROCEDURE remove(name IN VARCHAR2,
		value IN VARCHAR2,
		path IN VARCHAR2 DEFAULT NULL) AS
		string_cookies VARCHAR2(4000);					
	BEGIN
--	    string_cookies:='Set-Cookie:'||name||'='||value||';'||'expires='||to_char(SYSDATE-2000,'Day MON DD HH:MI:SS YYYY');
--	    IF path IS NOT NULL THEN
--			string_cookies:=string_cookies||';'||'path='||path;
--	    END IF;	
--	    oci_www.putStringHdr(string_cookies);
	    send(name,value,SYSDATE-1000,path);
	END; 
------------------------------
    PROCEDURE get_all(	    names out VC_ARR,
			    vals out VC_ARR,
			    num_vals out INTEGER) AS    
	header VARCHAR2(4000);
	header_left VARCHAR2(4000);
	search_name NUMBER;
	search_value NUMBER;
	k INTEGER;
	BEGIN
	    names:=VC_ARR();
	    vals:=VC_ARR();
	    num_vals:=0;
	    header:=';'||' '||OWA_UTIL.GET_CGI_ENV('http_cookie');
	    LOOP
		search_name:=INSTR(header,'=',1,(num_vals+1));
		header_left:=SUBSTR(header,search_name-1);
		EXIT WHEN search_name=0 OR search_name IS NULL;
		names.EXTEND;
		k:=names.LAST;
		names(k):=LTRIM(RTRIM(header,header_left),';');
		header:=header_left;
		search_value:=INSTR(header,';',1,(num_vals+1));
		header_left:=SUBSTR(header,search_value-1);
		vals.EXTEND;
		k:=vals.LAST;
		vals(k):=LTRIM(RTRIM(header,header_left),'=');
		num_vals:=num_vals+1;
	    END LOOP;
	END;    	
-------------------------------
    FUNCTION get(name IN VARCHAR2) RETURN cookie IS
		header VARCHAR2(4000);
		str	VARCHAR2(2000);
		old_pos NUMBER:=1;
		cur_pos NUMBER:=0;
		eq_pos	NUMBER;
		k INTEGER;
		cook COOKIE;
		
	BEGIN
		cook.num_vals:=0;
		cook.vals := VC_ARR();
		header:=OWA_UTIL.GET_CGI_ENV('http_cookie');
		LOOP
			cur_pos:=INSTR(header,';',old_pos,cook.num_vals+1);
			IF cur_pos = 0 THEN
				str:=SUBSTR(header,old_pos,LENGTH(header)-old_pos+1);
				cur_pos:=NULL;
			ELSE
				str:=SUBSTR(header,old_pos,cur_pos-old_pos);
			END IF;
			EXIT WHEN str IS NULL;
			eq_pos:=INSTR(str,'=');
			IF LTRIM(RTRIM(SUBSTR(str,1,eq_pos-1))) = name THEN
			    cook.vals.EXTEND;
			    k:=cook.vals.LAST;
			    cook.vals(k):=LTRIM(RTRIM(SUBSTR(str,eq_pos+1)));
			    cook.num_vals:=cook.num_vals+1;
			END IF;
			old_pos:=cur_pos+1;
		END LOOP;
	    RETURN cook;
	END;
END;
/
show err

drop public synonym owa_cookie;
create public synonym owa_cookie for owa_cookie;
grant execute on owa_cookie to public;
